from splight_models.constants import ComponentSize, RestartPolicy, LogginLevel, ComponentStatus
from splight_models.asset import Asset
from splight_models.attribute import Attribute
from splight_models.base import SplightBaseModel
from splight_models.file import File
from splight_models.datalake import ComponentDatalakeModel
from splight_models.graph import Graph
from splight_models.storage import StorageFile
from splight_models.query import Query
from splight_models.mapping import Mapping
from splight_models import EventActions, EventNames, CommunicationEvent
from datetime import datetime
from enum import Enum
from typing import Type, List, Dict, Tuple, Optional, Any, Union
from pydantic import BaseModel, create_model, Field, AnyUrl
from copy import copy
from functools import cached_property
import inspect


class Parameter(SplightBaseModel):
    name: str
    description: str = ''
    type: str = "str"
    required: bool = False
    multiple: bool = False
    sensitive: bool = False
    choices: Optional[List[Any]] = None
    depends_on: Optional[str] = None
    value: Any = None


class InputParameter(Parameter):
    pass


class OutputParameter(SplightBaseModel):
    name: str
    description: str = ''
    type: str
    choices: Optional[List[Any]] = None
    depends_on: Optional[str] = None
    filterable: bool = False


class CommandParameter(Parameter):
    pass


class CustomType(SplightBaseModel):
    _reserved_names = ["id", "name", "description"]
    name: str
    fields: List[Parameter]


class Output(SplightBaseModel):
    name: str
    fields: List[OutputParameter]


class Command(SplightBaseModel):
    name: str
    fields: List[CommandParameter]


class Binding(SplightBaseModel):
    name: str
    object_type: str
    object_action: EventActions


class ComponentObject(SplightBaseModel):
    id: Optional[str]
    component_id: str
    name: str
    description: Optional[str]
    type: str
    data: List[Parameter]

    @staticmethod
    def get_event_name(type: str, action: EventActions) -> str:
        return f"{type.lower()}_{action}"


class ComponentCommandResponse(SplightBaseModel):
    return_value: Optional[str] = None
    error_detail: Optional[str] = None


class ComponentCommandStatus(str, Enum):
    NOT_SENT = "not_sent"
    PENDING = "pending"
    SUCCESS = "success"
    ERROR = "error"


class ComponentCommand(SplightBaseModel):
    id: Optional[str]
    command: Command
    status: ComponentCommandStatus
    response: ComponentCommandResponse = ComponentCommandResponse()

    def get_event_name(self, action: EventActions) -> str:
        return f"componentcommand_{action}"


class ComponentCommandTriggerEvent(CommunicationEvent):
    event_name: str = Field(EventNames.COMPONENT_COMMAND_TRIGGER, const=True)
    data: ComponentCommand


class ComponentCommandCreateEvent(CommunicationEvent):
    event_name: str = Field(EventNames.COMPONENT_COMMAND_CREATE, const=True)
    data: ComponentCommand


class ComponentCommandUpdateEvent(CommunicationEvent):
    event_name: str = Field(EventNames.COMPONENT_COMMAND_UPDATE, const=True)
    data: ComponentCommand


class BaseComponent(SplightBaseModel):
    name: Optional[str] = None
    version: str
    custom_types: Optional[List[CustomType]] = []
    input: Optional[List[InputParameter]] = []
    output: Optional[List[Output]] = []
    commands: Optional[List[Command]] = []
    bindings: Optional[List[Binding]] = []

    class Config:
        keep_untouched = (cached_property,)

    @cached_property
    def custom_types_model(self) -> Type:
        return ComponentModelsFactory().get_custom_types_model(self.custom_types)

    @cached_property
    def input_model(self) -> Type:
        custom_type_model = self.custom_types_model
        custom_types = inspect.getmembers(custom_type_model)
        custom_types_dict = {a[0]: a[1] for a in custom_types if not a[0].startswith('__')}
        return ComponentModelsFactory(custom_types_dict).get_input_model(self.input)

    @cached_property
    def output_model(self) -> Type:
        return ComponentModelsFactory().get_output_model(self.output)

    @cached_property
    def commands_model(self) -> Type:
        custom_type_model = self.custom_types_model
        custom_types = inspect.getmembers(custom_type_model)
        custom_types_dict = {a[0]: a[1] for a in custom_types if not a[0].startswith('__')}
        return ComponentModelsFactory(custom_types_dict).get_commands_model(self.commands)


class Component(BaseComponent):
    id: Optional[str]
    description: Optional[str]
    log_level: LogginLevel = LogginLevel.info
    component_capacity: ComponentSize = ComponentSize.small
    restart_policy: RestartPolicy = RestartPolicy.ON_FAILURE
    status: ComponentStatus = ComponentStatus.STOPPED
    active: bool = False
    type: str = "Component"
    @property
    def collection(self):
        return str(self.id)


NATIVE_TYPES = {
    "int": int,
    "bool": bool,
    "str": str,
    "float": float,
    "datetime": datetime,
    "url": AnyUrl,
}

DATABASE_TYPES = {
    "Component": Component,
    "Asset": Asset,
    "Attribute": Attribute,
    "File": File,
    "Mapping": Mapping,
    "Graph": Graph,
    "Query": Query,
}

STORAGE_TYPES = {
    "file": StorageFile,
}

SIMPLE_TYPES = list(NATIVE_TYPES.keys()) + list(DATABASE_TYPES.keys()) + list(STORAGE_TYPES.keys())


class ComponentModelsFactory:
    def __init__(self, type_map: Dict[str, Type] = {}) -> None:
        self._type_map = {
            **type_map,
            **self._load_type_map()
        }

    @staticmethod
    def _load_type_map() -> Dict[str, Type]:
        type_map: Dict[str, Type] = {}
        type_map.update(NATIVE_TYPES)
        type_map.update({k: Union[str, v] for k, v in DATABASE_TYPES.items()})
        type_map.update({k: Union[str, v] for k, v in STORAGE_TYPES.items()})
        return type_map

    def get_input_model(self, inputs: List) -> BaseModel:
        # There is only one input model. No need to define a dict
        return self._create_model("Input", inputs)

    def get_custom_types_model(self, custom_types: List) -> Type:
        custom_types_dict: Dict[str, BaseModel] = {}
        for custom_type in custom_types:
            custom_type.fields.extend(
                [
                    Parameter(name=key, value=None)
                    for key in CustomType._reserved_names
                ]
            )
            model = self._create_model(custom_type.name, custom_type.fields)
            custom_types_dict[custom_type.name] = model
            self._type_map[custom_type.name] = model

        return type("CustomTypes", (), custom_types_dict)

    def get_output_model(self, outputs: List) -> BaseModel:
        output_models: Dict[str, BaseModel] = {}

        for output in outputs:
            output_format_field = {
                "output_format": Field(output.name, const=True),
            }
            output_models[output.name] = self._create_model(output.name,
                                                            output.fields,
                                                            output_format_field,
                                                            ComponentDatalakeModel)

        return type("Output", (), output_models)

    def get_commands_model(self, commands: List) -> BaseModel:
        command_models: Dict[str, BaseModel] = {}
        for command in commands:
            command_models[command.name] = self._create_model(command.name,
                                                              command.fields)
        return type("Commands", (), command_models)

    def _create_model(self,
                      name: str,
                      fields: List[Parameter],
                      extra_fields: Dict = {},
                      base: Type = SplightBaseModel) -> Type:
        class Meta:
            pass

        fields_dict: Dict[str, Tuple] = copy(extra_fields)
        for field in fields:
            setattr(Meta, field.name, field)
            type = self._type_map[field.type]
            choices = getattr(field, "choices", None)
            multiple = getattr(field, "multiple", False)
            required = getattr(field, "required", True)

            if choices:
                type = Enum(f"{field.name}-choices", {str(p): p for p in field.choices})

            if multiple:
                type = List[type]

            value = ...
            if not required:
                type = Optional[type]
                value = None

            fields_dict[field.name] = (type, value)

        model = create_model(
            name, **fields_dict, __base__=base
        )
        model.Meta = Meta
        return model



import pandas as pd
from pydantic import Field
from typing import Dict, Union, Optional
from .asset import Asset
from .attribute import Attribute
from .datalake import DatalakeModel, ComponentDatalakeModel


class Variable(DatalakeModel):
    instance_id: Optional[str] = None
    instance_type: Optional[str] = None
    args: Optional[Dict] = None
    path: Optional[str] = None
    asset_id: Optional[str] = None
    attribute_id: Optional[str] = None


class ConnectorOutput(ComponentDatalakeModel):
    asset: Union[Asset, str]
    attribute: Union[Attribute, str]


class Number(ConnectorOutput):
    output_format: str = Field("Number", const=True)
    value: float


class String(ConnectorOutput):
    output_format: str = Field("String", const=True)
    value: str


class Boolean(ConnectorOutput):
    output_format: str = Field("Boolean", const=True)
    value: bool


class VariableDataFrame(pd.DataFrame):
    pass



from abc import abstractmethod
from splight_models import Notification
from splight_abstract.client import AbstractClient


class AbstractNotificationClient(AbstractClient):
    @abstractmethod
    def send(self, instance: Notification, topic: str = 'default') -> None:
        self._client.trigger(self.channel, topic, instance.dict())



from typing import Optional
from .base import SplightBaseModel


class Graph(SplightBaseModel):
    id: Optional[str]
    title: str
    description: Optional[str] = None
    locked: bool = False


class Node(SplightBaseModel):
    id: Optional[str]
    type: str
    graph_id: str
    asset_id: Optional[str] = None
    color: str
    fill_color: Optional[str]
    position_x: int
    position_y: int
    width: str
    height: str

class Edge(SplightBaseModel):
    id: Optional[str]
    directed: bool
    graph_id: str
    asset_id: Optional[str] = None
    source_id: str
    target_id: str
    color: str
    source_handle: str
    target_handle: str



# TODO MOVE THIS STUFF
from typing import Type, Optional
from pydantic import BaseSettings, Field
from importlib import import_module
import os
import sys
from typing import Dict
TESTING = "test" in sys.argv or "pytest" in sys.argv
SPLIGHT_HOME = os.path.join(os.getenv('HOME'), '.splight')
USE_TZ = True
# EOTODO


"""
Settings for SPLIGHT framework are all namespaced in the SPLIGHT_FRAMEWORK setting.
For example your project's `settings.py` file might look like this:

SPLIGHT_FRAMEWORK = {
    'DATABASE_CLIENT': 'splight_lib.database.FakeDatabaseClient'
}

This module provides the `setup` object, that is used to access
SPLIGHT framework settings, checking for user settings first, then falling
back to the defaults.
"""


class SplightBaseSettings(BaseSettings):
    # TODO move all the private use only clients to private lib. Not need to define here those without remote implementation
    AUTH_CLIENT: str = 'fake_splight_lib.auth.FakeAuthClient'
    BLOCKCHAIN_CLIENT: str = 'fake_splight_lib.blockchain.FakeBlockchainClient'
    DATABASE_CLIENT: str = 'fake_splight_lib.database.FakeDatabaseClient'
    DATALAKE_CLIENT: str = 'fake_splight_lib.datalake.FakeDatalakeClient'
    DEPLOYMENT_CLIENT: str = 'fake_splight_lib.deployment.FakeDeploymentClient'
    CACHE_CLIENT: str = 'fake_splight_lib.cache.FakeCacheClient'
    COMMUNICATION_CLIENT: str = 'fake_splight_lib.communication.FakeCommunicationClient'
    NOTIFICATION_CLIENT: str = 'fake_splight_lib.notification.FakeNotificationClient'  # TODO deprecate this
    STORAGE_CLIENT: str = 'fake_splight_lib.storage.FakeStorageClient'
    HUB_CLIENT: str = 'fake_splight_lib.hub.FakeHubClient'
    COMPONENT_ID: Optional[str] = Field(None, env=["COMPONENT_ID"])
    NAMESPACE: str = "NO_NAMESPACE"

    @property
    def importables(self):
        return [
            'AUTH_CLIENT',
            'BILLING_CLIENT',
            'BLOCKCHAIN_CLIENT',
            'DATABASE_CLIENT',
            'DATALAKE_CLIENT',
            'DEPLOYMENT_CLIENT',
            'CACHE_CLIENT',
            'NOTIFICATION_CLIENT',
            'STORAGE_CLIENT',
            'HUB_CLIENT',
            'COMMUNICATION_CLIENT',
        ]


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError('Module "%s" does not define a "%s" attribute/class' % (
            module_path, class_name)
        ) from err


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, str):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        return import_string(val)
    except ImportError as e:
        msg = "Could not import '%s' for API setting '%s'. %s: %s." % (val, setting_name, e.__class__.__name__, e)
        raise ImportError(msg)


class SplightSettings:

    def __init__(self, user_settings={}, base_settings_model: Type[SplightBaseSettings] = SplightBaseSettings):
        self._cached_attrs = set()
        self._base_settings_model = base_settings_model
        self._base_settings = self._base_settings_model()
        self.configure(user_settings)

    @property
    def settings(self) -> Type[SplightBaseSettings]:
        return self._base_settings

    def __getattr__(self, attr):
        if not hasattr(self._base_settings, attr):
            raise AttributeError("Invalid API setting: '%s'" % attr)
        val = getattr(self._base_settings, attr)

        # Coerce import strings into classes
        if attr in self._base_settings.importables:
            val = perform_import(val, attr)

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def configure(self, user_settings: Dict):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        # This is setting the environment and
        # each model reloads the respective setup from there
        for key, value in user_settings.items():
            os.environ[key] = value
        # Reload settings
        self._base_settings = self._base_settings_model()


setup = SplightSettings()



from collections import UserList
from pydantic import BaseModel
from typing import List, Type, Callable
from functools import wraps
from abc import ABC
from splight_lib import logging
from .filter import FilterMixin
from .hooks import HooksMixin


logger = logging.getLogger()


class empty:
    pass


def validate_resource_type(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(self, resource_type: Type, *args, **kwargs):
        if resource_type not in self.valid_classes:
            raise NotImplementedError(f"Not a valid resource_type: {resource_type.__name__}")
        return func(self, resource_type, *args, **kwargs)
    return wrapper


def validate_instance_type(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(self, instance: BaseModel, *args, **kwargs):
        if type(instance) not in self.valid_classes:
            raise NotImplementedError(f"Not a valid instance type: {type(instance).__name__}")
        return func(self, instance, *args, **kwargs)
    return wrapper


class AbstractClient(ABC, HooksMixin, FilterMixin):
    valid_classes: List[Type] = []

    def __init__(self, namespace: str = "default", *args, **kwargs):
        self.namespace = namespace.lower().replace("_", "")

    def _validated_kwargs(self, resource_type: Type, **kwargs):
        '''
        Validate the given kwargs.
        '''
        class_fields = list(resource_type.__fields__.keys())
        return super()._validated_kwargs(class_fields, **kwargs)


class QuerySet(UserList):
    def __init__(self,
                 client: AbstractClient,
                 *args,
                 get_func: str = '_get',
                 count_func: str = 'count',
                 **kwargs):
        super().__init__()
        self._client = client
        self._client_func = get_func
        self._count_func = count_func
        self._args = args
        self._kwargs = kwargs
        self._cached_results = empty

    def __new__(cls, *args, **kwargs):
        obj = super(QuerySet, cls).__new__(cls)
        obj.__init__(*args, **kwargs)
        if kwargs.get('first', False):
            return obj.data
        return obj

    def filter(self, **kwargs) -> "QuerySet":
        kwargs = {**self._kwargs, **kwargs}
        return QuerySet(self._client, *self._args, **kwargs)

    def __repr__(self):
        data = [repr(obj) for obj in self[0:4]]
        extra = ", ..." if len(data) > 3 else ""
        return f"QuerySet([{', '.join(data[:3])}{extra}])"

    def __getitem__(self, i):
        if self._cached_results is not empty:
            return self._cached_results[i]

        if isinstance(i, slice):
            skip_ = self._kwargs.get("skip_", 0) + i.start

            limit_ = i.stop - i.start
            if "limit_" in self._kwargs:
                old_limit_ = self._kwargs["limit_"]
                limit_ = min(limit_, old_limit_)

            kwargs = {
                **self._kwargs
            }
            kwargs["skip_"] = skip_
            kwargs["limit_"] = limit_
            return QuerySet(self._client, get_func=self._client_func, count_func=self._count_func, *self._args, **kwargs)

        return self.data[i]

    def __len__(self):
        if self._cached_results is not empty:
            return len(self._cached_results)

        if hasattr(self._client, self._count_func):
            return getattr(self._client, self._count_func)(*self._args, **self._kwargs)

        return len(self.data)

    def count(self):
        return len(self)

    @property
    def data(self):
        if self._cached_results is empty:
            client_func = getattr(self._client, self._client_func)
            self._cached_results = client_func(*self._args, **self._kwargs)
        return self._cached_results

    @data.setter
    def data(self, value):
        self._cached_results = value



from abc import abstractmethod
from pydantic import BaseModel
from typing import Type, List
from splight_abstract.client import AbstractClient, QuerySet
from tempfile import TemporaryFile


class AbstractDatabaseClient(AbstractClient):

    @abstractmethod
    def save(self, instance: BaseModel) -> BaseModel:
        pass

    @abstractmethod
    def _get(self,
             resource_type: Type,
             first: bool = False,
             limit_: int = -1,
             skip_: int = 0,
             deleted: bool = False,
             **kwargs) -> List[BaseModel]:
        pass

    def get(self, resource_type: Type, *args, **kwargs) -> QuerySet:
        return QuerySet(self, resource_type, *args, **kwargs)

    @abstractmethod
    def delete(self, resource_type: Type, id: str) -> None:
        pass

    @abstractmethod
    def count(self, resource_type: Type, **kwargs) -> int:
        pass

    @abstractmethod
    def download(self, instance: BaseModel) -> TemporaryFile:
        pass



import uuid
from typing import List, Optional, Type

from pydantic import BaseModel

from splight_abstract import (
    AbstractDeploymentClient,
    validate_instance_type,
    validate_resource_type,
)
from splight_lib import logging
from splight_models import Deployment, Namespace

logger = logging.getLogger()


class FakeDeploymentClient(AbstractDeploymentClient):
    namespace = "default"
    deployments = {}
    valid_classes = [Deployment]

    def _get_deployment(self, id: str = "") -> List[Deployment]:
        logger.info("[FAKED] Retrieving fake deployment")
        if id:
            return [
                value for key, value in self.deployments.items() if key == id
            ]
        else:
            return [value for _, value in self.deployments.items()]

    @validate_resource_type
    def _get(
        self,
        resource_type: Type,
        first: bool = False,
        id: str = "",
        limit_: int = -1,
        skip_: int = 0,
        **kwargs,
    ) -> List[BaseModel]:
        if resource_type == Deployment:
            queryset = self._get_deployment(id=id)

        kwargs = self._validated_kwargs(resource_type, **kwargs)
        queryset = self._filter(queryset, **kwargs)
        if limit_ != -1:
            queryset = queryset[skip_:skip_ + limit_]

        if first:
            return queryset[0] if queryset else None
        return queryset

    def _list(self, **kwargs):
        queryset = self._get_deployment()
        return queryset

    def save(self, instance: BaseModel) -> BaseModel:
        raise NotImplementedError

    def delete(self, resource_type: Type, id: str) -> None:
        raise NotImplementedError

    def get_deployment_logs(self, id: str, limit: Optional[int] = None, since: Optional[str] = None, previous: bool = False) -> List[str]:
        raise NotImplementedError

    def get_capacity_options(self):
        raise NotImplementedError

    @classmethod
    def verify_header(cls, payload: bytes, signature: str) -> None:
        pass



from abc import abstractmethod
from pydantic import BaseModel
from typing import Type, List, Dict, Union, Callable
from datetime import timezone, timedelta
from splight_models import DatalakeModel, Query
from splight_abstract.client import AbstractClient, QuerySet
from functools import wraps
import pandas as pd


def validate_resource_type(func: Callable) -> Callable:
    @wraps(func)
    def inner(self, resource_type, *args, **kwargs):
        if not issubclass(resource_type, DatalakeModel):
            raise NotImplementedError(f"Not a valid resource_type: {resource_type.__name__}")
        return func(self, resource_type, *args, **kwargs)
    return inner


class AbstractDatalakeClient(AbstractClient):

    valid_filters = ["in", "contains", "gte", "lte"]

    @abstractmethod
    def save(self, instances: List[BaseModel], collection: str = "default") -> List[BaseModel]:
        pass

    @abstractmethod
    def raw_save(self, instances: List[Dict], collection: str = "default") -> Dict:
        pass

    def get(self, *args, **kwargs) -> QuerySet:
        return QuerySet(self, count_func="None", *args, **kwargs)

    def raw_get(self, *args, **kwargs) -> QuerySet:
        return QuerySet(self, get_func="_raw_get", count_func="None", *args, **kwargs)

    @abstractmethod
    def _get(self,
             resource_type: Type,
             collection: str = 'default',
             limit_: int = 50,
             skip_: int = 0,
             sort: Union[List, str] = ['timestamp__desc'],
             group_id: Union[List, str] = [],
             group_fields: Union[List, str] = [],
             tzinfo: timezone = timezone(timedelta()),
             **kwargs) -> List[BaseModel]:
        pass

    @abstractmethod
    def _raw_get(self,
                 collection: str = 'default',
                 limit_: int = 50,
                 skip_: int = 0,
                 sort: Union[List, str] = ['timestamp__desc'],
                 group_id: Union[List, str] = [],
                 group_fields: Union[List, str] = [],
                 tzinfo: timezone = timezone(timedelta()),
                 **kwargs) -> List[BaseModel]:
        pass

    @abstractmethod
    def get_dataframe(self, freq="H", collection: str = "default", **kwargs) -> pd.DataFrame:
        pass

    @abstractmethod
    def save_dataframe(self, dataframe: pd.DataFrame, collection: str = 'default') -> None:
        pass

    @abstractmethod
    def list_collection_names(self) -> List[str]:
        pass

    @abstractmethod
    def get_unique_keys(self, collection: str) -> List[str]:
        pass

    @abstractmethod
    def get_values_for_key(self, collection: str, key: str) -> List[str]:
        pass

    # Subject to incompatibility by implementation
    @abstractmethod
    def raw_aggregate(self, collection: str, pipeline: List[Dict]) -> List[Dict]:
        pass

    def _validated_kwargs(self, resource_type: Type, **kwargs):
        valid_fields: List[str] = list(resource_type.__fields__.keys())

        valid_filter: Dict[str, str] = {
            key: value for key, value in kwargs.items()
            if key in valid_fields or
            any(f"{valid_field}__" in key for valid_field in valid_fields)
        }

        return valid_filter

    @abstractmethod
    def create_index(self, collection: str, index: list) -> None:
        pass

    @abstractmethod
    def get_output(self, query: Query) -> List[Dict]:
        pass


from datetime import datetime
from typing import Optional

from splight_models.base import SplightBaseModel


class Credential(SplightBaseModel):
    access_id: str
    secret_key: Optional[str] = None
    created_date: Optional[datetime] = None
    last_used: Optional[datetime] = None
    user: Optional[int] = None
    description: Optional[str] = None



from abc import abstractmethod, ABC
from hexbytes import HexBytes
from splight_models.blockchain import CallResponse, Transaction, BlockchainContract


class AbstractBlockchainClient(ABC):

    @abstractmethod
    def get_balance(self) -> int:
        pass

    @abstractmethod
    def call(self, method: str, *args) -> CallResponse:
        pass

    @abstractmethod
    def transact(self, method: str, contract: BlockchainContract, *args, **kwargs) -> Transaction:
        pass

    @abstractmethod
    def get_transaction(self, tx_hast: HexBytes) -> Transaction:
        pass



from uuid import uuid4
from datetime import datetime, timezone
from typing import Dict, List, Type, Tuple

from pydantic import BaseModel
from splight_abstract.hub.abstract import AbstractHubSubClient
from splight_lib import logging
from splight_models import (
    HubComponent,
    HubComponentVersion,
)
from splight_abstract import AbstractHubClient, validate_resource_type


logger = logging.getLogger()


class FakeHubSubClient(AbstractHubSubClient):
    components = [
        HubComponent(id="1", name='Net1', description=None, version='01', input=[], splight_cli_version="0.1.0"),
        HubComponent(id="2", name='Net2', description=None, version='01', input=[], splight_cli_version="0.1.0"),
        HubComponent(id="3", name='Net3', description=None, version='01', input=[], splight_cli_version="0.1.0"),
        HubComponent(id="4", name='Algo1', description=None, version='01', input=[], splight_cli_version="0.1.0"),
        HubComponent(id="5", name='Algo2', description=None, version='01', input=[], splight_cli_version="0.1.0"),
        HubComponent(id="6", name='Conn1', description=None, version='01', input=[], splight_cli_version="0.1.0"),
        HubComponent(
            id="7",
            name="System1",
            description=None,
            version="01",
            input=[],
            splight_cli_version="0.1.0",
        )
    ]
    versions = components
    grouped_versions = components
    database: Dict[Type, List[BaseModel]] = {
        HubComponent: grouped_versions,
        HubComponentVersion: versions,
    }
    valid_classes = [
        HubComponent,
        HubComponentVersion
    ]

    allowed_update_fields = ["verification"]

    def save(self, instance: BaseModel) -> BaseModel:
        raise NotImplementedError

    @validate_resource_type
    def _get(self,
             resource_type: Type,
             first=False,
             limit_: int = -1,
             skip_: int = 0,
             **kwargs) -> List[BaseModel]:
        logger.debug(f"[FAKED] Pulling from hub {resource_type}s")
        queryset = self.database[resource_type]
        kwargs = self._validated_kwargs(resource_type, **kwargs)
        queryset = self._filter(queryset, **kwargs)
        if limit_ != -1:
            queryset = queryset[skip_:skip_ + limit_]

        if first:
            return queryset[0] if queryset else None

        return queryset

    @validate_resource_type
    def delete(self, resource_type: Type, id: str) -> None:
        raise NotImplementedError

    @validate_resource_type
    def update(self, resource_type: Type, id: str, data: Dict) -> BaseModel:
        instance = self.get(resource_type, id=id, first=True)
        if not instance:
            raise ValueError(f"{resource_type} with id {id} not found")
        for field in self.allowed_update_fields:
            if field in data:
                setattr(instance, field, data[field])
        return instance


class FakeHubClient(AbstractHubClient):

    def __init__(self, *args, **kwargs) -> None:
        self._client = FakeHubSubClient()

    def upload(self, data: Dict, files: Dict) -> Tuple:
        component = {
            "id": uuid4(),
            "tenant": "org_agu2n52305",
            "name": data["name"],
            "version": data["version"],
            "splight_cli_version": data["splight_cli_version"],
            "privacy_policy": data["privacy_policy"],
            "verification": data["verification"],
            "created_at": datetime.now(timezone.utc).isoformat(),
            "last_modified": datetime.now(timezone.utc).isoformat(),
            "readme": "some readme",
            "picture": "picture",
            "file": "file",
            "custom_types": [],
            "input": [],
            "output": [],
            "commands": [],
            "deleted": False
        }
        return component, 201

    def download(self, data: Dict) -> Tuple:
        return b"file content", 200

    def random_picture(self) -> Tuple:
        return b"image", 200

    @property
    def all(self) -> AbstractHubSubClient:
        return self._client

    @property
    def mine(self) -> AbstractHubSubClient:
        return self._client

    @property
    def public(self) -> AbstractHubSubClient:
        return self._client

    @property
    def private(self) -> AbstractHubSubClient:
        return self._client

    @property
    def setup(self) -> AbstractHubSubClient:
        return self._client

    @property
    def system(self) -> AbstractHubSubClient:
        return self._client
